public inbox for gcc-cvs@sourceware.org
help / color / mirror / Atom feed
* [gcc r12-1153] c++: value-init vs zero-init in expand_aggr_init_1 [PR65816]
@ 2021-06-01 20:21 Patrick Palka
  0 siblings, 0 replies; only message in thread
From: Patrick Palka @ 2021-06-01 20:21 UTC (permalink / raw)
  To: gcc-cvs, libstdc++-cvs

https://gcc.gnu.org/g:ac0bc21bd634a334ba8f323c39a11f01dfdc2aae

commit r12-1153-gac0bc21bd634a334ba8f323c39a11f01dfdc2aae
Author: Patrick Palka <ppalka@redhat.com>
Date:   Tue Jun 1 12:23:49 2021 -0400

    c++: value-init vs zero-init in expand_aggr_init_1 [PR65816]
    
    In the case of value-initializing an object of class type T,
    [dcl.init.general]/8 says:
    
      - if T has either no default constructor ([class.default.ctor]) or
        a default constructor that is user-provided or deleted, then the
        object is default-initialized;
      - otherwise, the object is zero-initialized and ...  if T has a
        non-trivial default constructor, the object is default-initialized;
    
    But when determining whether to first zero-initialize the object,
    expand_aggr_init_1 incorrectly considers the user-providedness of _all_
    constructors rather than only that of the _default_ constructors.  This
    causes us to skip the zero-initialization step when the class type has a
    defaulted default constructor alongside a user-defined constructor.
    
    It seems the predicate type_has_non_user_provided_default_constructor
    accurately captures the above rule for when to first perform a
    zero-initialization during value-initialization, so this patch adjusts
    expand_aggr_init_1 to use this predicate instead.
    
            PR c++/65816
    
    gcc/cp/ChangeLog:
    
            * init.c (expand_aggr_init_1): Check
            type_has_non_user_provided_default_constructor instead of
            type_has_user_provided_constructor.
    
    gcc/testsuite/ChangeLog:
    
            * g++.dg/cpp0x/constexpr-delegating3.C: New test.
            * g++.dg/cpp0x/dc10.C: New test.
            * g++.dg/cpp0x/initlist-base4.C: New test.
            * g++.dg/cpp2a/constexpr-init22.C: New test.
    
    libstdc++-v3/ChangeLog:
    
            * testsuite/23_containers/deque/allocator/default_init.cc,
            testsuite/23_containers/forward_list/allocator/default_init.cc,
            testsuite/23_containers/list/allocator/default_init.cc,
            testsuite/23_containers/map/allocator/default_init.cc,
            testsuite/23_containers/set/allocator/default_init.cc,
            testsuite/23_containers/vector/allocator/default_init.cc,
            testsuite/23_containers/vector/bool/allocator/default_init.cc:
            Remove xfail.

Diff:
---
 gcc/cp/init.c                                      |  4 ++--
 gcc/testsuite/g++.dg/cpp0x/constexpr-delegating3.C | 10 +++++++++
 gcc/testsuite/g++.dg/cpp0x/dc10.C                  | 19 ++++++++++++++++
 gcc/testsuite/g++.dg/cpp0x/initlist-base4.C        | 26 ++++++++++++++++++++++
 gcc/testsuite/g++.dg/cpp2a/constexpr-init22.C      | 14 ++++++++++++
 .../23_containers/deque/allocator/default_init.cc  |  1 -
 .../forward_list/allocator/default_init.cc         |  1 -
 .../23_containers/list/allocator/default_init.cc   |  1 -
 .../23_containers/map/allocator/default_init.cc    |  1 -
 .../23_containers/set/allocator/default_init.cc    |  1 -
 .../23_containers/vector/allocator/default_init.cc |  1 -
 .../vector/bool/allocator/default_init.cc          |  1 -
 12 files changed, 71 insertions(+), 9 deletions(-)

diff --git a/gcc/testsuite/g++.dg/cpp0x/constexpr-delegating3.C b/gcc/testsuite/g++.dg/cpp0x/constexpr-delegating3.C
new file mode 100644
index 00000000000..2263ec89488
--- /dev/null
+++ b/gcc/testsuite/g++.dg/cpp0x/constexpr-delegating3.C
@@ -0,0 +1,10 @@
+// PR c++/65816
+// { dg-do compile { target c++11 } }
+
+struct test {
+  int m;
+  test() = default;
+  constexpr test(int) : test() {}
+};
+
+static_assert(test(0).m == 0, "");
diff --git a/gcc/testsuite/g++.dg/cpp0x/dc10.C b/gcc/testsuite/g++.dg/cpp0x/dc10.C
new file mode 100644
index 00000000000..c008a1703e8
--- /dev/null
+++ b/gcc/testsuite/g++.dg/cpp0x/dc10.C
@@ -0,0 +1,19 @@
+// PR c++/65816
+// { dg-do run { target c++11 } }
+
+void* operator new(decltype(sizeof(int)), void* ptr) { return ptr; }
+
+struct test {
+  int i;
+  test() = default;
+  test(int) : test() {}
+};
+
+int main() {
+  alignas(test) unsigned char space[sizeof(test)];
+  for (auto& c : space) c = 0xff;
+
+  auto ptr = ::new(&space) test(42);
+  int& i = static_cast<test&>(*ptr).i;
+  if (i != 0) __builtin_abort();
+}
diff --git a/gcc/testsuite/g++.dg/cpp0x/initlist-base4.C b/gcc/testsuite/g++.dg/cpp0x/initlist-base4.C
new file mode 100644
index 00000000000..4a02af92799
--- /dev/null
+++ b/gcc/testsuite/g++.dg/cpp0x/initlist-base4.C
@@ -0,0 +1,26 @@
+// PR c++/65816
+// { dg-do run { target c++11 } }
+
+void* operator new(decltype(sizeof(int)), void* ptr) { return ptr; }
+
+struct item { int i; };
+
+struct collector : item {
+  int j;
+  collector() = default;
+  collector(int) {}
+};
+
+struct tuple : collector {
+  tuple() : collector() {}
+};
+
+int main() {
+  alignas(tuple) unsigned char space[sizeof(tuple)];
+  for (auto& c : space) c = 0xff;
+
+  auto ptr = ::new(&space) tuple;
+  int& i = static_cast<tuple&>(*ptr).i;
+  int& j = static_cast<tuple&>(*ptr).j;
+  if (i != 0 || j != 0) __builtin_abort();
+}
diff --git a/gcc/testsuite/g++.dg/cpp2a/constexpr-init22.C b/gcc/testsuite/g++.dg/cpp2a/constexpr-init22.C
new file mode 100644
index 00000000000..50e4a95dbda
--- /dev/null
+++ b/gcc/testsuite/g++.dg/cpp2a/constexpr-init22.C
@@ -0,0 +1,14 @@
+// PR c++/65816
+// { dg-do compile { target c++20 } }
+
+struct X {
+  int i;
+  X() = default;
+  constexpr X(int) { }
+};
+
+struct Y : X {
+  constexpr Y() : X() { }
+};
+
+static_assert(Y().i == 0);
diff --git a/libstdc++-v3/testsuite/23_containers/deque/allocator/default_init.cc b/libstdc++-v3/testsuite/23_containers/deque/allocator/default_init.cc
index f51cc5b49b9..9428206f2ef 100644
--- a/libstdc++-v3/testsuite/23_containers/deque/allocator/default_init.cc
+++ b/libstdc++-v3/testsuite/23_containers/deque/allocator/default_init.cc
@@ -17,7 +17,6 @@
 
 // { dg-do run { target c++11 } }
 // { dg-options "-O0" }
-// { dg-xfail-run-if "PR c++/65816" { *-*-* } }
 
 #include <deque>
 #include <testsuite_hooks.h>
diff --git a/libstdc++-v3/testsuite/23_containers/forward_list/allocator/default_init.cc b/libstdc++-v3/testsuite/23_containers/forward_list/allocator/default_init.cc
index a60192641d4..4298fa5ffa5 100644
--- a/libstdc++-v3/testsuite/23_containers/forward_list/allocator/default_init.cc
+++ b/libstdc++-v3/testsuite/23_containers/forward_list/allocator/default_init.cc
@@ -17,7 +17,6 @@
 
 // { dg-do run { target c++11 } }
 // { dg-options "-O0" }
-// { dg-xfail-run-if "PR c++/65816" { *-*-* } }
 
 #include <forward_list>
 #include <testsuite_hooks.h>
diff --git a/libstdc++-v3/testsuite/23_containers/list/allocator/default_init.cc b/libstdc++-v3/testsuite/23_containers/list/allocator/default_init.cc
index 62bfd78df18..03d8cfc89e8 100644
--- a/libstdc++-v3/testsuite/23_containers/list/allocator/default_init.cc
+++ b/libstdc++-v3/testsuite/23_containers/list/allocator/default_init.cc
@@ -17,7 +17,6 @@
 
 // { dg-do run { target c++11 } }
 // { dg-options "-O0" }
-// { dg-xfail-run-if "PR c++/65816" { *-*-* } }
 
 #include <list>
 #include <testsuite_hooks.h>
diff --git a/libstdc++-v3/testsuite/23_containers/map/allocator/default_init.cc b/libstdc++-v3/testsuite/23_containers/map/allocator/default_init.cc
index bc9027703fc..6e3f36d60a4 100644
--- a/libstdc++-v3/testsuite/23_containers/map/allocator/default_init.cc
+++ b/libstdc++-v3/testsuite/23_containers/map/allocator/default_init.cc
@@ -17,7 +17,6 @@
 
 // { dg-do run { target c++11 } }
 // { dg-options "-O0" }
-// { dg-xfail-run-if "PR c++/65816" { *-*-* } }
 
 #include <map>
 #include <testsuite_hooks.h>
diff --git a/libstdc++-v3/testsuite/23_containers/set/allocator/default_init.cc b/libstdc++-v3/testsuite/23_containers/set/allocator/default_init.cc
index 8a99b491c25..b031cb81504 100644
--- a/libstdc++-v3/testsuite/23_containers/set/allocator/default_init.cc
+++ b/libstdc++-v3/testsuite/23_containers/set/allocator/default_init.cc
@@ -17,7 +17,6 @@
 
 // { dg-do run { target c++11 } }
 // { dg-options "-O0" }
-// { dg-xfail-run-if "PR c++/65816" { *-*-* } }
 
 #include <set>
 #include <testsuite_hooks.h>
diff --git a/libstdc++-v3/testsuite/23_containers/vector/allocator/default_init.cc b/libstdc++-v3/testsuite/23_containers/vector/allocator/default_init.cc
index c16d3b046e4..4673db71b6c 100644
--- a/libstdc++-v3/testsuite/23_containers/vector/allocator/default_init.cc
+++ b/libstdc++-v3/testsuite/23_containers/vector/allocator/default_init.cc
@@ -17,7 +17,6 @@
 
 // { dg-do run { target c++11 } }
 // { dg-options "-O0" }
-// { dg-xfail-run-if "PR c++/65816" { *-*-* } }
 
 #include <vector>
 #include <testsuite_hooks.h>
diff --git a/libstdc++-v3/testsuite/23_containers/vector/bool/allocator/default_init.cc b/libstdc++-v3/testsuite/23_containers/vector/bool/allocator/default_init.cc
index 4111e3ac3eb..61cc7859d74 100644
--- a/libstdc++-v3/testsuite/23_containers/vector/bool/allocator/default_init.cc
+++ b/libstdc++-v3/testsuite/23_containers/vector/bool/allocator/default_init.cc
@@ -17,7 +17,6 @@
 
 // { dg-do run { target c++11 } }
 // { dg-options "-O0" }
-// { dg-xfail-run-if "PR c++/65816" { *-*-* } }
 
 #include <vector>
 #include <testsuite_hooks.h>


^ permalink raw reply	[flat|nested] only message in thread

only message in thread, other threads:[~2021-06-01 20:21 UTC | newest]

Thread overview: (only message) (download: mbox.gz / follow: Atom feed)
-- links below jump to the message on this page --
2021-06-01 20:21 [gcc r12-1153] c++: value-init vs zero-init in expand_aggr_init_1 [PR65816] Patrick Palka

This is a public inbox, see mirroring instructions
for how to clone and mirror all data and code used for this inbox;
as well as URLs for read-only IMAP folder(s) and NNTP newsgroup(s).
diff --git a/gcc/cp/init.c b/gcc/cp/init.c
index 04d495807ef..b1123287300 100644
--- a/gcc/cp/init.c
+++ b/gcc/cp/init.c
@@ -2078,9 +2078,9 @@ expand_aggr_init_1 (tree binfo, tree true_exp, tree exp, tree init, int flags,
      that's value-initialization.  */
   if (init == void_type_node)
     {
       /* If the type has data but no user-provided ctor, we need to zero
 	 out the object.  */
-      if (!type_has_user_provided_constructor (type)
+      if (type_has_non_user_provided_default_constructor (type)
 	  && !is_really_empty_class (type, /*ignore_vptr*/true))
 	{
 	  tree field_size = NULL_TREE;
@@ -2078,9 +2078,9 @@ expand_aggr_init_1 (tree binfo, tree true_exp, tree exp, tree init, int flags,
   if (init == void_type_node)
     {
-      /* If the type has data but no user-provided ctor, we need to zero
+      /* If the type has data but no user-provided default ctor, we need to zero
 	 out the object.  */